@@ -6,23 +6,38 @@ NUM_STEPS=0
 export PB_POSITION=0
 export PB_WAYPOINT_LENGTH=0
 
-# used by kernel installation code
+# used for setting up apt
 PROTOCOL=
 MIRROR=
 DIRECTORY=
 COMPONENTS=
 DISTRIBUTION=
-INCLUDES=
-EXCLUDES=
+
+# used by kernel installation code
 KERNEL=
 KERNEL_LIST=/tmp/available_kernels.txt
-KERNEL_MAJOR="$(uname -r | cut -d . -f 1,2)"
+KERNEL_NAME=`udpkg --print-os`
+case $KERNEL_NAME in
+	linux)		KERNEL_MAJOR="$(uname -r | cut -d - -f 1 | cut -d . -f 1,2)" ;;
+	kfreebsd)	KERNEL_MAJOR="$(uname -r | cut -d . -f 1)" ;;
+	hurd)		KERNEL_NAME=gnumach ; KERNEL_MAJOR="$(uname -v | cut -d ' ' -f 2 | cut -d . -f 1)" ;;
+esac
 KERNEL_VERSION="$(uname -r | cut -d - -f 1)"
 KERNEL_ABI="$(uname -r | cut -d - -f 1,2)"
+KERNEL_FLAVOUR=$(uname -r | cut -d - -f 3-)
 MACHINE="$(uname -m)"
 NUMCPUS=$(cat /var/numcpus 2>/dev/null) || true
 CPUINFO=/proc/cpuinfo
-SPEAKUP=/proc/speakup
+
+# files and directories
+APT_SOURCES=/target/etc/apt/sources.list
+APT_CONFDIR=/target/etc/apt/apt.conf.d
+IT_CONFDIR=/target/etc/initramfs-tools/conf.d
+DPKG_CONFDIR=/target/etc/dpkg/dpkg.cfg.d
+
+IFS_ORIG="$IFS"
+NL="
+"
 
 log() {
 	logger -t base-installer "$@"
@@ -66,7 +81,7 @@ run_waypoints () {
 
 update_progress () {
 	# Updates the progress bar to a new position within the space allocated
-	# for the current waypoint. 
+	# for the current waypoint.
 	NW_POS=$(($PB_POSITION + $PB_WAYPOINT_LENGTH * $1 / $2))
 	db_progress SET $NW_POS
 }
@@ -74,11 +89,12 @@ update_progress () {
 check_target () {
 	# Make sure something is mounted on the target.
 	# Partconf causes the latter format.
-	if ! grep -q '/target ' /proc/mounts && \
+	if [ -e /proc/mounts ] && \
+	   ! grep -q '/target ' /proc/mounts && \
 	   ! grep -q '/target/ ' /proc/mounts; then
 		exit_error base-installer/no_target_mounted
 	fi
-	
+
 	# Warn about installation over an existing unix.
 	if [ -e /target/bin/sh -o -L /target/bin/sh ]; then
 		warning "attempting to install to unclean target"
@@ -106,7 +122,7 @@ check_target () {
 	fi
 }
 
-setup_dev () {
+setup_dev_linux () {
 	# Ensure static device nodes created during install are preserved
 	# Tests in MAKEDEV require this is done in the D-I environment
 	mkdir -p /dev/.static/dev
@@ -114,82 +130,100 @@ setup_dev () {
 	mount --bind /target/dev /dev/.static/dev
 	# Mirror device nodes in D-I environment to target
 	mount --bind /dev /target/dev/
+}
 
-	# RAID
-	if [ -e /proc/mdstat ] && grep -q ^md /proc/mdstat ; then
-		apt-install mdadm
-	fi
-	# device-mapper
-	if grep -q " device-mapper$" /proc/misc; then
-		# Avoid warnings from lvm2 tools about open file descriptors
-		export LVM_SUPPRESS_FD_WARNINGS=1
+setup_dev_kfreebsd() {
+	mount -t devfs devfs /target/dev
+}
 
-		# We can't check the root node directly as is done above because
-		# root could be on an LVM LV on top of an encrypted device
-		if type dmsetup >/dev/null 2>&1 && \
-		   dmsetup table | cut -d' ' -f4 | grep -q "crypt" 2>/dev/null; then
-			apt-install cryptsetup
-		fi
-
-		if type dmraid >/dev/null 2>&1; then
-			if [ "$(dmraid -s -c | grep -v "No RAID disks")" ]; then
-				apt-install dmraid
-			fi
-		fi
-
-		if pvdisplay | grep -iq "physical volume ---"; then
-			apt-install lvm2
-		fi
-	fi
+setup_dev() {
+	case "$OS" in
+		linux) setup_dev_linux ;;
+		kfreebsd) setup_dev_kfreebsd ;;
+		hurd) : ;;
+		*) warning "setup_dev called for an unknown OS ($OS)." ;;
+	esac	
 }
 
 configure_apt_preferences () {
-	[ ! -d /target/etc/apt/apt.conf.d ] && mkdir -p /target/etc/apt/apt.conf.d
-	
+	[ ! -d "$APT_CONFDIR" ] && mkdir -p "$APT_CONFDIR"
+
+	# Don't install Recommends during base-installer
+	cat >$APT_CONFDIR/00InstallRecommends <<EOT
+APT::Install-Recommends "false";
+EOT
+
 	# Make apt trust Debian CDs. This is not on by default (we think).
 	# This will be left in place on the installed system.
-	cat > /target/etc/apt/apt.conf.d/00trustcdrom <<EOT
+	cat > $APT_CONFDIR/00trustcdrom <<EOT
 APT::Authentication::TrustCDROM "true";
 EOT
 
 	# Avoid clock skew causing gpg verification issues.
 	# This file will be left in place until the end of the install.
-	cat > /target/etc/apt/apt.conf.d/00IgnoreTimeConflict << EOT
+	cat > $APT_CONFDIR/00IgnoreTimeConflict << EOT
 Acquire::gpgv::Options { "--ignore-time-conflict"; };
 EOT
 
 	if db_get debian-installer/allow_unauthenticated && [ "$RET" = true ]; then
 		# This file will be left in place until the end of the install.
-		cat > /target/etc/apt/apt.conf.d/00AllowUnauthenticated << EOT
+		cat > $APT_CONFDIR/00AllowUnauthenticated << EOT
 APT::Get::AllowUnauthenticated "true";
 Aptitude::CmdLine::Ignore-Trust-Violations "true";
 EOT
 	fi
+
+	if [ "$PROTOCOL" = https ] && db_get debian-installer/allow_unauthenticated_ssl && [ "$RET" = true ]; then
+		# This file will be left in place on the installed system.
+		cat > $APT_CONFDIR/00AllowUnauthenticatedSSL << EOT
+Acquire::https::Verify-Host "false";
+Acquire::https::Verify-Peer "false";
+EOT
+	fi
+
+	# Disable all syncing; it's unnecessary in an installation context,
+	# and can slow things down quite a bit.
+	# This file will be left in place until the end of the install.
+	cat > $DPKG_CONFDIR/force-unsafe-io <<EOT
+force-unsafe-io
+EOT
+}
+
+final_apt_preferences () {
+	# From here on install Recommends as configured
+	db_get base-installer/install-recommends
+	if [ "$RET" = true ]; then
+		rm -f $APT_CONFDIR/00InstallRecommends
+	fi
 }
 
 apt_update () {
 	log-output -t base-installer chroot /target apt-get update \
 		|| apt_update_failed=$?
-	
+
 	if [ "$apt_update_failed" ]; then
 		warning "apt update failed: $apt_update_failed"
 	fi
 }
 
 install_extra () {
+	local IFS
 	info "Installing queued packages into /target/."
-	
-	if [ -f /var/lib/apt-install/queue ] ; then
+
+	if [ -f /var/lib/apt-install/queue ]; then
 		# We need to install these one by one in case one fails.
-		PKG_COUNT=`cat /var/lib/apt-install/queue | wc -w`
+		PKG_COUNT=$(cat /var/lib/apt-install/queue | wc -w)
 		CURR_PKG=0
-		for PKG in `cat /var/lib/apt-install/queue`; do
+		IFS="$NL"
+		for LINE in $(cat /var/lib/apt-install/queue); do
+			IFS="$IFS_ORIG"
+			PKG=${LINE%% *}
+			OPTS=$(echo "$LINE" | sed "s/$PKG *//")
 			db_subst base-installer/section/install_extra_package SUBST0 "$PKG"
 			db_progress INFO base-installer/section/install_extra_package
 
-			if ! log-output -t base-installer apt-install $PKG; then
+			log-output -t base-installer apt-install $OPTS $PKG || \
 				warning "Failed to install $PKG into /target/: $?"
-			fi
 
 			# Advance progress bar within space allocated for install_extra
 			CURR_PKG=$(($CURR_PKG + 1))
@@ -199,15 +233,18 @@ install_extra () {
 }
 
 pre_install_hooks () {
+	# avoid apt-install installing things; apt is not configured yet
+	rm -f $APT_SOURCES
+
 	partsdir="/usr/lib/base-installer.d"
 	if [ -d "$partsdir" ]; then
-		for script in `ls "$partsdir"/*`; do
+		for script in `ls "$partsdir"/* 2>/dev/null`; do
 			base=$(basename $script | sed 's/[0-9]*//')
 			if ! db_progress INFO base-installer/progress/$base; then
 				db_subst base-installer/progress/fallback SCRIPT "$base"
 				db_progress INFO base-installer/progress/fallback
 		    	fi
-	
+
 		    	if [ -x "$script" ] ; then
 				# be careful to preserve exit code
 				if log-output -t base-installer "$script"; then
@@ -223,9 +260,12 @@ pre_install_hooks () {
 }
 
 post_install_hooks () {
+	# locales will now be installed, so unset
+	unset IT_LANG_OVERRIDE
+
 	partsdir="/usr/lib/post-base-installer.d"
 	if [ -d "$partsdir" ]; then
-		scriptcount=`ls "$partsdir"/* | wc -l`
+		scriptcount=`ls "$partsdir"/* 2>/dev/null | wc -l`
 		scriptcur=0
 		for script in "$partsdir"/*; do
 			base="$(basename "$script" | sed 's/[0-9]*//')"
@@ -272,16 +312,9 @@ get_mirror_info () {
 		else
 			COMPONENTS="*"
 		fi
-		if [ -s /cdrom/.disk/base_include ]; then
-			INCLUDES=`grep -v '^#' /cdrom/.disk/base_include | tr '\n' , | sed 's/,$//'`
-		fi
-			
-		if [ -s /cdrom/.disk/base_exclude ]; then
-			EXCLUDES=`grep -v '^#' /cdrom/.disk/base_exclude | tr '\n' , | sed 's/,$//'`
-		fi
 
-		# Sanity check: an error reading /cdrom/.disk/base_components can cause
-		# ugly errors in debootstrap because $COMPONENTS will be empty.
+		# Sanity check: an error reading /cdrom/.disk/base_components can
+		# cause ugly errors in debootstrap because $COMPONENTS will be empty
 		if [ -z "$COMPONENTS" ]; then
 			exit_error base-installer/cannot_install
 		fi
@@ -293,7 +326,7 @@ get_mirror_info () {
 		fi
 
 		mirror_error=""
-		
+
 		db_get mirror/protocol || mirror_error=1
 		PROTOCOL="$RET"
 
@@ -320,9 +353,14 @@ kernel_update_list () {
 	(set +e;
 	# Hack to get the metapackages in the right order; should be
 	# replaced by something better at some point.
-	chroot /target apt-cache search ^linux- | grep '^linux-\(amd64\|386\|686\|k7\|generic\|server\|virtual\|rt\|xen\|power\|cell\|itanium\|mckinley\|sparc\|hppa\)';
-	chroot /target apt-cache search ^linux-image- | grep -v '^linux-image-2\.';
-	chroot /target apt-cache search ^linux-image-2. | sort -r) | \
+	chroot /target apt-cache search ^linux-signed- | grep '^linux-signed-\(generic\|server\|virtual\|preempt\|rt\|xen\)';
+	chroot /target apt-cache search ^linux- | grep '^linux-\(amd64\|686\|k7\|generic\|server\|virtual\|preempt\|rt\|xen\|power\|cell\|ia64\|sparc\|hppa\|imx51\|dove\|omap\|omap4\|armadaxp\|highbank\)';
+	chroot /target apt-cache search ^linux-signed-image- | grep -v '^linux-signed-image-[2-9]\.';
+	chroot /target apt-cache search ^linux-image- | grep -v '^linux-image-[2-9]\.';
+	chroot /target apt-cache search '^linux-signed-image-[2-9]\.' | sort -r;
+	chroot /target apt-cache search '^linux-image-[2-9]\.' | sort -r;
+	chroot /target apt-cache search ^kfreebsd-image;
+	chroot /target apt-cache search ^gnumach-image) | \
 	cut -d" " -f1 | uniq > "$KERNEL_LIST.unfiltered"
 	kernels=`< "$KERNEL_LIST.unfiltered" tr '\n' ' ' | sed -e 's/ $//'`
 	for candidate in $kernels; do
@@ -346,6 +384,8 @@ kernel_present () {
 
 pick_kernel () {
 	kernel_update_list
+	
+	db_settitle debian-installer/bootstrap-base/title
 
 	# Check for overrides
 	if db_get base-installer/kernel/override-image && [ "$RET" ]; then
@@ -392,11 +432,19 @@ pick_kernel () {
 		db_fset base-installer/kernel/image seen false || true
 
 		if [ -n "$FLAVOUR" ]; then
-			arch_kernel=$(arch_get_kernel "$FLAVOUR")
+			arch_kernel="$(arch_get_kernel "$FLAVOUR")"
+
+			# Hack to support selection of meta packages with a postfix
+			# added to the normal name (for updated kernels in stable).
+			if db_get base-installer/kernel/altmeta && [ "$RET" ]; then
+				arch_kernel="$(echo "$arch_kernel" | \
+					sed "s/$/-$RET/"; \
+					echo "$arch_kernel")"
+			fi
 		else
 			arch_kernel=""
 		fi
-	
+
 		got_arch_kernel=
 		if [ "$arch_kernel" ]; then
 			info "arch_kernel candidates: $arch_kernel"
@@ -445,15 +493,20 @@ pick_kernel () {
 	info "Using kernel '$KERNEL'"
 }
 
-install_linux () {
+install_kernel_linux () {
 	if [ "$KERNEL" = none ]; then
 		info "Not installing any kernel"
 		return
 	fi
-	
-	target_kernel_major="$(echo "$KERNEL" | sed 's/^kernel-image-//; s/^linux-image-//; s/-.*//' | cut -d . -f 1,2)"
+
+	target_kernel_major="$(echo "$KERNEL" | sed 's/^kernel-image-//; s/^linux-\(\|signed-\)image-//; s/-.*//' | cut -d . -f 1,2)"
 	case $target_kernel_major in
 		2.?)	;;
+		[3-9].*)
+			# As far as our debconf templates are concerned,
+			# this is essentially 2.6.
+			target_kernel_major=2.6
+			;;
 		*)
 			# something went wrong; use major version of
 			# installation kernel
@@ -476,6 +529,10 @@ install_linux () {
 		do_initrd=yes
 	fi
 
+	if [ `archdetect` = mipsel/loongson-2f ]; then
+	    do_initrd=yes
+	fi
+
 	if db_get base-installer/kernel/linux/link_in_boot ; then
 		if [ "$RET" = "true" ]; then
 			link_in_boot=yes
@@ -499,32 +556,20 @@ install_linux () {
 # Kernel image management overrides
 # See kernel-img.conf(5) for details
 do_symlinks = yes
-relative_links = yes
 do_bootloader = no
-do_bootfloppy = no
 do_initrd = $do_initrd
 link_in_boot = $link_in_boot
 EOF
 
-	# TODO This should probably be restructured to better support
-	#      differences between initrd generators
 	rd_generator=""
 	if [ "$do_initrd" = yes ]; then
-		if generators="$(available_initramfs_generators)"; then
-			if echo "$generators" | grep -q " "; then
-				rd_generator="$(select_initramfs_generator "$generators")"
-			else
-				rd_generator="$generators"
-			fi
-		fi
+		rd_generator=initramfs-tools
 
 		# initramfs-tools needs busybox-initramfs pre-installed (and
 		# only recommends it)
-		if [ "$rd_generator" = "initramfs-tools" ]; then
-			if ! log-output -t base-installer apt-install busybox-initramfs; then
-				db_subst base-installer/kernel/failed-package-install PACKAGE busybox-initramfs
-				exit_error base-installer/kernel/failed-package-install
-			fi
+		if ! log-output -t base-installer apt-install busybox-initramfs; then
+			db_subst base-installer/kernel/failed-package-install PACKAGE busybox-initramfs
+			exit_error base-installer/kernel/failed-package-install
 		fi
 
 		# Make sure the ramdisk creation tool is installed before we
@@ -535,17 +580,6 @@ EOF
 			db_subst base-installer/kernel/failed-package-install PACKAGE "$rd_generator"
 			exit_error base-installer/kernel/failed-package-install
 		fi
-		
-		# Figure out how to configure the ramdisk creation tool
-		# FJP 20070306: Possibly this can go completely
-		case "$rd_generator" in
-		    initramfs-tools|yaird)
-			: ;;
-		    *)
-			db_subst base-installer/initramfs/unsupported GENERATOR "$rd_generator"
-			exit_error base-installer/initramfs/unsupported
-			;;
-		esac
 
 		# Add modules that have been queued for inclusion in the initrd
 		FIRSTMODULE=1
@@ -553,10 +587,33 @@ EOF
 			[ ! -e $QUEUEFILE ] && break
 			MODULE=$(basename $QUEUEFILE ".initrd")
 			addmodule_initramfs_tools "$MODULE" $FIRSTMODULE
-			addmodule_yaird "$MODULE" $FIRSTMODULE
 			rm $QUEUEFILE
 			FIRSTMODULE=0
 		done
+
+		# Select and set driver inclusion policy for initramfs-tools
+		if db_get base-installer/initramfs-tools/driver-policy && \
+		   [ -z "$RET" ]; then
+			# Get default for architecture
+			db_get base-installer/kernel/linux/initramfs-tools/driver-policy
+			db_set base-installer/initramfs-tools/driver-policy "$RET"
+		fi
+		db_settitle debian-installer/bootstrap-base/title
+		db_input medium base-installer/initramfs-tools/driver-policy || true
+		if ! db_go; then
+			db_progress stop
+			exit 10
+		fi
+
+		db_get base-installer/initramfs-tools/driver-policy
+		if [ "$RET" != most ]; then
+			cat > $IT_CONFDIR/driver-policy <<EOF
+# Driver inclusion policy selected during installation
+# Note: this setting overrides the value set in the file
+# /etc/initramfs-tools/initramfs.conf
+MODULES=$RET
+EOF
+		fi
 	else
 		info "Not installing an initrd generator."
 	fi
@@ -581,7 +638,7 @@ EOF
 		# Set up a default resume partition.
 		case $rd_generator in
 		    initramfs-tools)
-			resumeconf=/target/etc/initramfs-tools/conf.d/resume
+			resumeconf=$IT_CONFDIR/resume
 			;;
 		    *)
 			resumeconf=
@@ -593,8 +650,8 @@ EOF
 		else
 			resume=
 		fi
-		if [ "$resume" ] && PATH="/lib/udev:$PATH" type vol_id >/dev/null 2>&1; then
-			resume_uuid="$(PATH="/lib/udev:$PATH" vol_id -u "$resume" || true)"
+		if [ "$resume" ] && ! echo "$resume" | grep -q "^/dev/mapper/"; then
+			resume_uuid="$(block-attr --uuid "$resume" || true)"
 			if [ "$resume_uuid" ]; then
 				resume="UUID=$resume_uuid"
 			fi
@@ -611,7 +668,7 @@ EOF
 		# Set PReP root partition
 		if [ "$ARCH" = powerpc ] && [ "$SUBARCH" = prep ] && \
 		   [ "$rd_generator" = initramfs-tools ]; then
-			prepconf=/target/etc/initramfs-tools/conf.d/prep-root
+			prepconf=$IT_CONFDIR/prep-root
 			rootpart_devfs=$(mount | grep "on /target " | cut -d' ' -f1)
 			rootpart=$(mapdevfs $rootpart_devfs)
 			if [ -f $prepconf ] && grep -q "^#* *ROOT=" $prepconf; then
@@ -623,7 +680,7 @@ EOF
 		fi
 	fi
 
-	# Advance progress bar to 30% of allocated space for install_linux
+	# Advance progress bar to 30% of allocated space for install_kernel_linux
 	update_progress 30 100
 
 	# Install the kernel
@@ -631,6 +688,34 @@ EOF
 	db_progress INFO base-installer/section/install_kernel_package
 	log-output -t base-installer apt-install "$KERNEL" || kernel_install_failed=$?
 
+	db_get base-installer/kernel/headers
+	if [ "$RET" = true ]; then
+		# Advance progress bar to 80% of allocated space for install_kernel_linux
+		update_progress 80 100
+
+		# Install kernel headers if possible
+		HEADERS="$(echo "$KERNEL" | sed 's/linux\(-image\|\)/linux-headers/')"
+		db_subst base-installer/section/install_kernel_package SUBST0 "$HEADERS"
+		db_progress INFO base-installer/section/install_kernel_package
+		log-output -t base-installer apt-install "$HEADERS" || true
+	fi
+
+	db_get base-installer/kernel/backports-modules
+	if [ "$RET" ]; then
+		BACKPORTS_MODULES="$RET"
+
+		# Advance progress bar to 85% of allocated space for install_kernel_linux
+		update_progress 85 100
+
+		# Install kernel backports modules if possible
+		for backports_module in $BACKPORTS_MODULES; do
+			LBM="$(echo "$KERNEL" | sed "s/linux\\(-image\\|\\)/linux-backports-modules-$backports_module-$DISTRIBUTION/")"
+			db_subst base-installer/section/install_kernel_package SUBST0 "$LBM"
+			db_progress INFO base-installer/section/install_kernel_package
+			log-output -t base-installer apt-install "$LBM" || true
+		done
+	fi
+
 	# Advance progress bar to 90% of allocated space for install_linux
 	update_progress 90 100
 
@@ -652,6 +737,9 @@ get_resume_partition () {
 		if [ "$type" != partition ]; then
 			continue
 		fi
+		if [ ! -e "$filename" ]; then
+			continue
+		fi
 		if [ "${filename#/dev/ramzswap}" != "$filename" ]; then
 			continue
 		fi
@@ -663,43 +751,6 @@ get_resume_partition () {
 	echo "$biggest_partition"
 }
 
-available_initramfs_generators () {
-	irf_list=""
-	db_get base-installer/kernel/linux/initramfs-generators || return 1
-
-	for irf in $RET; do
-		if LANG=C chroot /target apt-cache policy $irf 2>&1 | \
-			grep "Candidate:" | grep -v "(none)" >/dev/null 2>&1; then
-			if [ "$irf_list" ]; then
-				irf_list="$irf_list $irf"
-			else
-				irf_list="$irf"
-			fi
-		fi
-	done
-	info "Available initramfs generator(s): '$irf_list'"
-
-	[ "$irf_list" ] || return 1
-	echo "$irf_list"
-	return 0
-}
-
-select_initramfs_generator () {
-	irf_choices="$(echo "$1" | sed "s/ \+/, /g")"
-	irf_default="${1%% *}"
-	db_subst base-installer/initramfs/generator GENERATORS "$irf_choices"
-	db_set base-installer/initramfs/generator "$irf_default"
-
-	db_input low base-installer/initramfs/generator || [ $? -eq 30 ]
-	if ! db_go; then
-		db_progress stop
-		exit 10
-	fi
-
-	db_get base-installer/initramfs/generator
-	echo $RET
-}
-
 addmodule_easy () {
 	if [ -f "$CFILE" ]; then
 		if [ "$2" = 1 ]; then
@@ -714,70 +765,182 @@ addmodule_initramfs_tools () {
 	addmodule_easy "$1" "$2"
 }
 
-addmodule_yaird () {
-	CFILE='/target/etc/yaird/Default.cfg'
-	if [ -f "$CFILE" ]; then
-		if [ "$2" = 1 ]; then
-			sed -i "/END GOALS/s/^/\n\t\t#\n\t\t# Added by Debian Installer\n/" $CFILE
-		fi
-		sed -i "/END GOALS/s/^/\t\tMODULE $1\n/" $CFILE
+install_kernel_kfreebsd() {
+	if [ "$KERNEL" = none ]; then
+		info "Not installing any kernel"
+		return
+	fi
+
+	# Create configuration file for kernel-package
+	if [ -f /target/etc/kernel-img.conf ]; then
+		# Backup old kernel-img.conf
+		mv /target/etc/kernel-img.conf /target/etc/kernel-img.conf.$$
+	fi
+
+	cat > /target/etc/kernel-img.conf <<EOF
+# Kernel image management overrides
+# See kernel-img.conf(5) for details
+do_symlinks = no
+EOF
+	# Advance progress bar to 10% of allocated space for install_kfreebsd
+	update_progress 10 100
+
+	# Install the kernel
+	db_subst base-installer/section/install_kernel_package SUBST0 "$KERNEL"
+	db_progress INFO base-installer/section/install_kernel_package
+	log-output -t base-installer apt-install "$KERNEL" || kernel_install_failed=$?
+
+	# Advance progress bar to 90% of allocated space for install_kfreebsd
+	update_progress 90 100
+
+	if [ -f /target/etc/kernel-img.conf.$$ ]; then
+		# Revert old kernel-img.conf
+		mv /target/etc/kernel-img.conf.$$ /target/etc/kernel-img.conf
+	fi
+
+	if [ "$kernel_install_failed" ]; then
+		db_subst base-installer/kernel/failed-install KERNEL "$KERNEL"
+		exit_error base-installer/kernel/failed-install
 	fi
 }
 
-configure_apt () {
-	# let apt inside the chroot see the cdrom
-	if [ "$PROTOCOL" = file ] ; then
-		if [ -n "$DIRECTORY" ]; then
-			umount /target$DIRECTORY 2>/dev/null || true
-			if [ ! -e /target/$DIRECTORY ]; then
-				mkdir -p /target/$DIRECTORY
-			fi
-		fi
-		if ! mount -o bind $DIRECTORY /target$DIRECTORY; then
-			warning "failed to bind mount /target$DIRECTORY"
-		fi
-		# The bind mount is left mounted, for future apt-install
-		# calls to use.
+install_kernel_hurd() {
+	if [ "$KERNEL" = none ]; then
+		info "Not installing any kernel"
+		return
 	fi
 
-	# sources.list uses space to separate the components, not comma
-	COMPONENTS=`echo $COMPONENTS | tr , " "`
-	APTSOURCE="$PROTOCOL://$MIRROR$DIRECTORY"
+	# Create configuration file for kernel-package
+	if [ -f /target/etc/kernel-img.conf ]; then
+		# Backup old kernel-img.conf
+		mv /target/etc/kernel-img.conf /target/etc/kernel-img.conf.$$
+	fi
 
-	# This assumes the file protocol is only used for CD (image) installs
-	if [ "$PROTOCOL" = file ] ; then
+	cat > /target/etc/kernel-img.conf <<EOF
+# Kernel image management overrides
+# See kernel-img.conf(5) for details
+do_symlinks = no
+EOF
+	# Advance progress bar to 10% of allocated space for install_kfreebsd
+	update_progress 10 100
+
+	# Install the kernel
+	db_subst base-installer/section/install_kernel_package SUBST0 "$KERNEL"
+	db_progress INFO base-installer/section/install_kernel_package
+	log-output -t base-installer apt-install "$KERNEL" || kernel_install_failed=$?
+
+	# Advance progress bar to 90% of allocated space for install_kernel_hurd
+	update_progress 90 100
+
+	if [ -f /target/etc/kernel-img.conf.$$ ]; then
+		# Revert old kernel-img.conf
+		mv /target/etc/kernel-img.conf.$$ /target/etc/kernel-img.conf
+	fi
+
+	if [ "$kernel_install_failed" ]; then
+		db_subst base-installer/kernel/failed-install KERNEL "$KERNEL"
+		exit_error base-installer/kernel/failed-install
+	fi
+}
+
+install_kernel() {
+	case "$OS" in
+		linux) install_kernel_linux ;;
+		kfreebsd) install_kernel_kfreebsd ;;
+		hurd) install_kernel_hurd ;;
+		*) warning "install_kernel called for an unknown OS ($OS)." ;;
+	esac	
+}
+
+
+# Assumes the file protocol is only used for CD (image) installs
+configure_apt () {
+	if [ "$PROTOCOL" = file ]; then
+		local tdir=/target/media$DIRECTORY
+		rm -f /var/lib/install-cd.id
+
+		# Let apt inside the chroot see the cdrom
+		umount $tdir 2>/dev/null || true
+		if [ ! -e $tdir ]; then
+			mkdir -p $tdir
+		fi
+
+		# The bind mount is left mounted, for future apt-install
+		# calls to use.
+		case "$OS" in
+			linux)
+			if ! mount -o bind $DIRECTORY $tdir; then
+				warning "failed to bind mount $tdir"
+			fi
+			;;
+			kfreebsd)
+			if ! mount -t nullfs $DIRECTORY $tdir ; then
+				warning "failed to bind mount $tdir"
+			fi
+			;;
+			hurd)
+			if ! mount -t firmlink $DIRECTORY $tdir > /dev/null 2>&1 ; then
+				warning "failed to bind mount $tdir"
+			fi
+			;;
+			*)
+			warning "configure_apt called with unknown OS ($OS)."
+			;;
+		esac
+
+		# Define the mount point for apt-cdrom
+		cat > $APT_CONFDIR/00CDMountPoint << EOT
+Acquire::cdrom {
+  mount "/media/cdrom";
+};
+Dir::Media::MountPath "/media/cdrom";
+EOT
 		# Make apt-cdrom and apt not unmount/mount CD-ROMs;
-		# needed to support CD images (hd-media installs)
-		# This file will be left in place until the end of the install.
-		cat > /target/etc/apt/apt.conf.d/00NoMountCDROM << EOT
+		# needed to support CD images (hd-media installs).
+		# This file will be left in place until the end of the
+		# install for hd-media installs, but is removed again
+		# during apt-setup for installs using real CD/DVDs.
+		cat > $APT_CONFDIR/00NoMountCDROM << EOT
 APT::CDROM::NoMount "true";
 Acquire::cdrom {
-  mount "/cdrom";
-  "/cdrom/" {
+  "/media/cdrom/" {
     Mount  "true";
     UMount "true";
   };
+  AutoDetect "false";
 }
 EOT
+
 		# Scan CD-ROM or CD image; start with clean sources.list
-		: > /target/etc/apt/sources.list
-		if ! log-output -t base-installer chroot /target apt-cdrom add ; then
+		# Prevent apt-cdrom from prompting
+		: > $APT_SOURCES
+		if ! log-output -t base-installer \
+		     chroot /target apt-cdrom add </dev/null; then
 			error "error while running apt-cdrom"
 		fi
 		if db_get apt-setup/restricted && [ "$RET" = false ]; then
-			sed -i 's/ \(restricted\|multiverse\)//g' /target/etc/apt/sources.list
+			sed -i 's/ \(restricted\|multiverse\)//g' $APT_SOURCES
 		fi
 	else
-		echo "deb $APTSOURCE $DISTRIBUTION $COMPONENTS" > /target/etc/apt/sources.list
-		echo "deb $APTSOURCE $DISTRIBUTION-updates $COMPONENTS" >> /target/etc/apt/sources.list
+		# sources.list uses space to separate the components, not comma
+		COMPONENTS=$(echo $COMPONENTS | tr , " ")
+		APTSOURCE="$PROTOCOL://$MIRROR$DIRECTORY"
+
+		echo "deb $APTSOURCE $DISTRIBUTION $COMPONENTS" > $APT_SOURCES
+		echo "deb $APTSOURCE $DISTRIBUTION-updates $COMPONENTS" >> $APT_SOURCES
 		if db_get apt-setup/security_host; then
 			SECMIRROR="$RET"
 		else
 			SECMIRROR="$MIRROR"
 		fi
-		echo "deb $PROTOCOL://$SECMIRROR/ubuntu $DISTRIBUTION-security $COMPONENTS" >> /target/etc/apt/sources.list
+		if db_get apt-setup/security_path; then
+			SECDIRECTORY="$RET"
+		else
+			SECDIRECTORY=/ubuntu
+		fi
+		echo "deb $PROTOCOL://$SECMIRROR$SECDIRECTORY $DISTRIBUTION-security $COMPONENTS" >> $APT_SOURCES
 		if db_get apt-setup/proposed && [ "$RET" = true ]; then
-			echo "deb $APTSOURCE $DISTRIBUTION-proposed $COMPONENTS" >> /target/etc/apt/sources.list
+			echo "deb $APTSOURCE $DISTRIBUTION-proposed $COMPONENTS" >> $APT_SOURCES
 		fi
 	fi
 }
